/*
 * NOTE: This example is based on the example in the kernel source's 
 * samples/kprobes/kprobe_example.c file - I took out code relating to other 
 * processor platforms, as well as adding additional comments to explain things.
 * You can view the original in your own kernel source code.
 *
 * This sample kernel module showing the use of kprobes to dump a
 * stack trace and selected registers when _do_fork() is called.
 *
 * For more information on theory of operation of kprobes, see
 * Documentation/kprobes.txt
 *
 * You will see the trace data in /var/log/messages and on the console
 * whenever _do_fork() is invoked to create a new process.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kprobes.h>

/* For each probe you need to allocate a kprobe structure 
   symbol_name specifies where to insert your probe. You can alternatively
   also use .addr to specify an actual address you want to "hijack"
*/
static struct kprobe kp = {
	.symbol_name	= "_do_fork", 
};

/* kprobe pre_handler: called just before the probed instruction is executed */
static int handler_pre(struct kprobe *p, struct pt_regs *regs)
{
	printk(KERN_ALERT "pre_handler: p->addr = 0x%p, ip = %lx,"
			" flags = 0x%lx\n",
		p->addr, regs->ip, regs->flags);

	/* A dump_stack() here will give a stack backtrace */
	return 0;
}

/* kprobe post_handler: called after the probed instruction is executed */
static void handler_post(struct kprobe *p, struct pt_regs *regs,
				unsigned long flags)
{
	printk(KERN_ALERT "post_handler: p->addr = 0x%p, flags = 0x%lx\n",
		p->addr, regs->flags);
}

/*
 * fault_handler: this is called if an exception is generated for any
 * instruction within the pre- or post-handler, or when Kprobes
 * single-steps the probed instruction.
 */
static int handler_fault(struct kprobe *p, struct pt_regs *regs, int trapnr)
{
	printk(KERN_ALERT "fault_handler: p->addr = 0x%p, trap #%dn",
		p->addr, trapnr);
	/* Return 0 because we don't handle the fault. */
	return 0;
}

static int __init kprobe_init(void)
{
	int ret;
	kp.pre_handler = handler_pre;
	kp.post_handler = handler_post;
	kp.fault_handler = handler_fault;

    /*
     Register the kprobe. Note that the struct kprobe we pass to this function
     MUST remain allocated until you have called unregister_kprobe!
    */
	ret = register_kprobe(&kp); 
	if (ret < 0) {
		printk(KERN_ALERT "register_kprobe failed, returned %d\n", ret);
		return ret;
	}
	printk(KERN_ALERT "Planted kprobe at %p\n", kp.addr);
	return 0;
}

static void __exit kprobe_exit(void)
{
    //were kp dynamically allocated, it would be important to unregister first,
    //and only afterwards free the memory. Of course this is not the case here.
	unregister_kprobe(&kp);
	printk(KERN_ALERT "kprobe at %p unregistered\n", kp.addr);
}

module_init(kprobe_init)
module_exit(kprobe_exit)
MODULE_LICENSE("GPL");
